@@ -1,4 +1,4 @@
-# $Id: Kernel.pm,v 1.19 1998/11/23 19:59:45 troc Exp $
+# $Id: Kernel.pm,v 1.22 1998/11/26 17:56:17 troc Exp $
 # Documentation exists after __END__
 
 package POE::Kernel;
@@ -8,7 +8,12 @@ use POSIX qw(EINPROGRESS EINTR);
 use IO::Select;
 use Carp;
                                         # allow subsecond alarms, if available
-eval { require Time::HiRes; import Time::HiRes qw(time sleep); };
+BEGIN {
+  eval {
+    require Time::HiRes;
+    import Time::HiRes qw(time sleep);
+  };
+}
 
 #------------------------------------------------------------------------------
 # states  : [ [ $session, $source_session, $state, $time, \@etc ], ... ]
@@ -23,9 +28,11 @@ eval { require Time::HiRes; import Time::HiRes qw(time sleep); };
 # signals:  { $signal => { $session => $state, ... } }
 #
 # sessions: { $session => [ $parent, \@children, $states, $selects, $session,
-#                           $signals,
+#                           \%signals, \%names
 #                         ]
 #           }
+#
+# names: { $name => $session }
 #------------------------------------------------------------------------------
 
 # a list of signals that will terminate all sessions (and stop the kernel)
@@ -56,6 +63,7 @@ sub new {
                     'states'   => [ ],
                     'signals'  => { },
                     'terminal' => { },
+                    'names'    => { },
                    }, $type;
                                         # these can't be up in the main bless?
   $self->{'sel_r'} = new IO::Select() || die $!;
@@ -78,6 +86,30 @@ sub new {
   $self;
 }
 
+#------------------------------------------------------------------------------
+# check a session for death by starvation
+
+sub _collect_garbage {
+  my ($self, $session) = @_;
+                                        # check for death by starvation
+  if (($session ne $self) && (exists $self->{'sessions'}->{$session})) {
+
+#     warn( "***** $session  states(" . $self->{'sessions'}->{$session}->[2] .
+#           ")  selects(" . $self->{'sessions'}->{$session}->[3] .
+#           ")  children(" . scalar(@{$self->{'sessions'}->{$session}->[1]}) .
+#           ")\n"
+#         );
+
+    unless ($self->{'sessions'}->{$session}->[2] || # queued states
+            $self->{'sessions'}->{$session}->[3] || # pending selects
+            @{$self->{'sessions'}->{$session}->[1]} || # children
+            keys(%{$self->{'sessions'}->{$session}->[6]}) # registered names
+    ) {
+      $self->session_free($session);
+    }
+  }
+}
+
 #------------------------------------------------------------------------------
 # Send a state to a session right now.  Used by _disp_select to expedite
 # select() states, and used by run() to deliver posted states from the queue.
@@ -88,7 +120,7 @@ sub _dispatch_state {
                                         # add this session to kernel tables
   if ($state eq '_start') {
     $self->{'sessions'}->{$session} =
-      [ $source_session, [ ], 0, 0, $session, 0, ];
+      [ $source_session, [ ], 0, 0, $session, { }, { }, ];
     unless ($session eq $source_session) {
       push(@{$self->{'sessions'}->{$source_session}->[1]}, $session);
     }
@@ -104,6 +136,7 @@ sub _dispatch_state {
     }
                                         # tell the parent its child is gone
     $self->_dispatch_state($parent, $session, '_child', []);
+    $self->_collect_garbage($parent);
   }
                                         # dispatch signal to children
   elsif ($state eq '_signal') {
@@ -147,7 +180,7 @@ sub _dispatch_state {
     }
     $self->{'sessions'}->{$session}->[1] = [ ];
                                         # free lingering signals
-    foreach my $signal (@{$self->{'blocked signals'}}) {
+    foreach my $signal (keys(%{$self->{'sessions'}->{$session}->[5]})) {
       $self->_internal_sig($session, $signal);
     }
                                         # free lingering states (if leaking?)
@@ -163,6 +196,11 @@ sub _dispatch_state {
     foreach my $handle (@handles) {
       $self->_kernel_select($session, $self->{'selects'}->{$handle}->[3]);
     }
+                                        # free lingering names
+    foreach my $name (keys(%{$self->{'sessions'}->{$session}->[6]})) {
+      delete $self->{'names'}->{$name};
+    }
+    $self->{'sessions'}->{$session}->[6] = { };
                                         # check for leaks -><- debugging only
     if (my $leaked = @{$self->{'sessions'}->{$session}->[1]}) {
       print "*** $session - leaking children ($leaked)\n";
@@ -173,8 +211,12 @@ sub _dispatch_state {
     if (my $leaked = $self->{'sessions'}->{$session}->[3]) {
       print "*** $session - leaking selects ($leaked)\n";
     }
-    if (my $leaked = $self->{'sessions'}->{$session}->[5]) {
+    # number 4 is the session itself, blessed
+    if (my $leaked = keys(%{$self->{'sessions'}->{$session}->[5]})) {
       print "*** $session - leaking signals ($leaked)\n";
+    }
+    if (my $leaked = keys(%{$self->{'sessions'}->{$session}->[6]})) {
+      print "*** $session - leaking names ($leaked)\n";
     }
                                         # remove this session (should be empty)
     delete $self->{'sessions'}->{$session};
@@ -189,22 +231,8 @@ sub _dispatch_state {
       $self->session_free($session);
     }
   }
-                                        # check for death by starvation
-  elsif (($session ne $self) && (exists $self->{'sessions'}->{$session})) {
-
-#     warn( "***** $session  states(" . $self->{'sessions'}->{$session}->[2] .
-#           ")  selects(" . $self->{'sessions'}->{$session}->[3] .
-#           ")  children(" . scalar(@{$self->{'sessions'}->{$session}->[1]}) .
-#           ")\n"
-#         );
-
-    unless ($self->{'sessions'}->{$session}->[2] || # queued states
-            $self->{'sessions'}->{$session}->[3] || # pending selects
-            @{$self->{'sessions'}->{$session}->[1]} # children
-    ) {
-      $self->session_free($session);
-    }
-  }
+                                        # return what the state handler did
+  $handled;
 }
 
 #------------------------------------------------------------------------------
@@ -221,6 +249,7 @@ sub _dispatch_selects {
       foreach my $notify (@selects) {
         my ($session, $state) = @$notify;
         $self->_dispatch_state($session, $session, $state, [ $handle ]);
+        $self->_collect_garbage($session);
       }
     }
     else {
@@ -273,6 +302,7 @@ sub run {
           = @{shift @{$self->{'states'}}};
         $self->{'sessions'}->{$session}->[2]--;
         $self->_dispatch_state($session, $source_session, $state, $etc);
+        $self->_collect_garbage($session);
       }
     }
   }
@@ -343,6 +373,7 @@ sub session_alloc {
 #  $self->_enqueue_state($session, $active_session, '_start', time(), []);
 
   $self->_dispatch_state($session, $active_session, '_start', []);
+  $self->_collect_garbage($session);
   $self->{'active session'} = $active_session;
 }
 
@@ -353,6 +384,7 @@ sub session_free {
     unless (exists $self->{'sessions'}->{$session});
 
   $self->_dispatch_state($session, $self->{'active session'}, '_stop', []);
+  $self->_collect_garbage($session);
 
 #  $self->_enqueue_state($session, $self->{'active session'},
 #                        '_stop', time(), []
@@ -454,15 +486,11 @@ sub _internal_sig {
   my ($self, $session, $signal, $state) = @_;
 
   if ($state) {
-    unless (exists $self->{'signals'}->{$signal}->{$session}) {
-      $self->{'sessions'}->{$session}->[5]++;
-    }
+    $self->{'sessions'}->{$session}->[5]->{$signal} = $state;
     $self->{'signals'}->{$signal}->{$session} = $state;
   }
   else {
-    if (exists $self->{'signals'}->{$signal}->{$session}) {
-      $self->{'sessions'}->{$session}->[5]--;
-    }
+    delete $self->{'sessions'}->{$session}->[5]->{$signal};
     delete $self->{'signals'}->{$signal}->{$session};
   }
 }
@@ -500,7 +528,7 @@ sub delay {
     $self->alarm($state, time() + $delay, @etc);
   }
   else {
-    $self->alarm($state, 0, @etc);
+    $self->alarm($state, 0);
   }
 }
 
@@ -550,22 +578,41 @@ sub sig {
   $self->_internal_sig($self->{'active session'}, $signal, $state);
 }
 
+sub signal {
+  my ($self, $session, $signal) = @_;
+  my $active_session = $self->{'active session'};
+  $session = $self->alias_resolve($session);
+  $self->_enqueue_state($session, $active_session,
+                        '_signal', time(), [ $signal ]
+                       );
+}
+
 #------------------------------------------------------------------------------
 # Post a state to the queue.
 
 sub post {
-  my ($self, $destination_session, $state_name, @etc) = @_;
+  my ($self, $destination, $state_name, @etc) = @_;
   my $active_session = $self->{'active session'};
-                                        # external -> internal representation
-  if ($destination_session eq $active_session->{'namespace'}) {
-    $destination_session = $active_session;
-  }
+  $destination = $self->alias_resolve($destination);
 
-  $self->_enqueue_state($destination_session, $active_session,
+  $self->_enqueue_state($destination, $active_session,
                         $state_name, time(), \@etc
                        );
 }
 
+#------------------------------------------------------------------------------
+# Call a state directly.
+
+sub call {
+  my ($self, $destination, $state_name, @etc) = @_;
+  my $active_session = $self->{'active session'};
+  $destination = $self->alias_resolve($destination);
+
+  $self->_dispatch_state($destination, $active_session,
+                         $state_name, \@etc
+                        );
+}
+
 #------------------------------------------------------------------------------
 # State management.
 
@@ -576,6 +623,58 @@ sub state {
   $active_session->register_state($state_name, $state_code);
 }
 
+#------------------------------------------------------------------------------
+# Alias management.
+
+sub alias_set {
+  my ($self, $name) = @_;
+  my $active_session = $self->{'active session'};
+
+  if (exists $self->{'names'}->{$name}) {
+    if ($self->{'names'}->{$name} ne $active_session) {
+      carp "alias '$name' already exists for a different session";
+      return 0;
+    }
+  }
+  else {
+    $self->{'names'}->{$name} = $active_session;
+    $self->{'sessions'}->{$active_session}->[6]->{$name} = 1;
+  }
+}
+
+sub alias_remove {
+  my ($self, $name) = @_;
+  my $active_session = $self->{'active session'};
+
+  if (exists $self->{'names'}->{$name}) {
+    if ($self->{'names'}->{$name} ne $active_session) {
+      carp "alias '$name' does not refer to the current session";
+      return 0;
+    }
+    delete $self->{'names'}->{$name};
+    delete $self->{'sessions'}->{$active_session}->[6]->{$name};
+  }
+}
+
+sub alias_resolve {
+  my ($self, $name) = @_;
+  my $resolved_session = $self->{'active session'};
+  if ($name ne $resolved_session->{'namespace'}) {
+    if (ref($name) eq '') {
+      if (exists $self->{'names'}->{$name}) {
+        $resolved_session = $self->{'names'}->{$name};
+      }
+      else {
+        $resolved_session = '';
+      }
+    }
+    else {
+      $resolved_session = $name;
+    }
+  }
+  $resolved_session;
+}
+
 ###############################################################################
 1;
 __END__
@@ -667,10 +766,28 @@ names as with C<%SIG>).  If C<$state> is defined, then that state will be
 invoked when a specified signal.  If C<$state> is undefined, then the
 C<$SIG{$signal}> handler is removed.
 
+=item $kernel->signal($session, $signal)
+
+Sends a signal to a session (and all child sessions).  C<$session> may
+also refer to a C<POE::Kernel> instance, in which case it will be
+propagated to every session managed by that kernel.
+
 =item $kernel->post($destination_session, $state_name, @etc)
 
-Enqueues an event (C<$state>) for the C<$destination_session>.  Additional
-parameters (C<@etc>) can be passed along.
+Enqueues an event (C<$state_name>) for the C<$destination_session>.
+Additional parameters (C<@etc>) can be passed along.
+
+C<$destination_session> may be an alias, or a Kernel instance.
+
+=item $kernel->call($destination_session, $state_name, @etc)
+
+Directly calls an event handler (C<$state_name>) for the
+C<$destination_session>.  Additional parameters (C<@etc>) can be
+passed along.
+
+Returns the event handler's return value.
+
+C<$destination_session> may be an alias, or a Kernel instance.
 
 =item $kernel->state($state_name, $state_code)
 
@@ -708,6 +825,33 @@ whichever C<time()> that POE::Kernel does.  Otherwise, POE::Session
 code may be using a different version of C<time()>, and subsecond
 delays may not be working as expected.
 
+=item $kernel->alias_set($alias)
+
+Sets a name (alias) for the current session.  This allows the session
+to be referenced by name instead of by Perl reference.  Sessions may
+have more than one alias.
+
+Sessions with one or more aliases are treated as daemons.  They will
+not be garbage-collected even if they have no states, selects or
+children.  This allows "service" sessions to be created without
+resorting to posting unnecessary "keep alive" events.
+
+An alias can only point to one Session per Kernel.  Aliases are case
+sensitive.
+
+Daemons will be garbage-collected during SIGZOMBIE processing, when
+there are no more events to process.
+
+=item $kernel->alias_remove($alias)
+
+Clears the name (alias) for the current session.  This "undaemonizes"
+the session.
+
+=item $object = $kernel->alias_resolve($name)
+
+Resolves an alias to the object it references.  Returns an empty
+string if the alias does not exist.
+
 =back
 
 =head1 PROTECTED METHODS
